from collections import defaultdict
from aio_pubsub.interfaces import PubSub, Subscriber
from queue import Queue
from weakref import WeakSet


class MemorySubscriber(Subscriber):
    def __init__(self, queue_factory):
        self.messages = queue_factory(maxsize=0)

    def __aiter__(self):
        return self

    async def __anext__(self):
        return self.messages.get()

    async def put(self, message):
        self.messages.put_nowait(message)


class MemoryPubSub(PubSub):
    def __init__(self, queue_factory=Queue):
        self.subscribers = defaultdict(WeakSet)
        self.queue_factory = queue_factory

    async def publish(self, channel, message):
        subscribers = self.subscribers[channel]
        for subscriber in subscribers:
            await subscriber.put(message)
        return len(subscribers)

    async def subscribe(self, channel):
        subscriber = MemorySubscriber(self.queue_factory)
        self.subscribers[channel].add(subscriber)
        return subscriber



import json
from aio_pubsub.interfaces import PubSub, Subscriber

aioredis_installed = False
try:
    import aioredis

    aioredis_installed = True
except ImportError:
    pass  # pragma: no cover


class RedisSubscriber(Subscriber):
    def __init__(self, channel):
        self.channel = channel

    def __aiter__(self):
        return self

    async def __anext__(self):
        async for msg in self.channel.iter(encoding="utf-8", decoder=json.loads):
            return msg


class RedisPubSub(PubSub):
    def __init__(self, dsn="redis://localhost:6379/0?encoding=utf-8", timeout=10):
        if aioredis_installed is False:
            raise RuntimeError("Please install `aioredis`")  # pragma: no cover
        self.dsn = dsn
        self.sub = None
        self.pub = None
        self.timeout = timeout

    async def init(self):
        self.sub = await aioredis.create_redis(self.dsn, timeout=self.timeout)
        self.pub = await aioredis.create_redis(self.dsn, timeout=self.timeout)

    async def publish(self, channel, message):
        channels = await self.pub.pubsub_channels(channel)
        for ch in channels:
            await self.pub.publish_json(ch, message)

    async def subscribe(self, channel):
        channel = await self.sub.subscribe(channel)
        return RedisSubscriber(channel[0])

    async def close(self):
        self.sub.close()
        self.pub.close()



import typing
from abc import ABCMeta, abstractmethod


class Subscriber(metaclass=ABCMeta):
    """
    Subscriber interface
    """

    @abstractmethod
    def __aiter__(self):
        """The subscriber iterable"""


class PubSub(metaclass=ABCMeta):
    """
    Main interface for implementation
    """

    @abstractmethod
    async def publish(self, channel: typing.Any, message: typing.Any):
        """Publish a message to channel"""

    @abstractmethod
    async def subscribe(self, channel: typing.Any):
        """Subscribe to one or many channels and return a Subscriber object"""



import asyncio
import logging
from datetime import datetime
from aio_pubsub.interfaces import PubSub, Subscriber

motor_installed = False
try:
    from pymongo.errors import CollectionInvalid
    from pymongo import CursorType
    import motor.motor_asyncio

    motor_installed = True
except ImportError:
    pass  # pragma: no cover


class MongoSubscriber(Subscriber):
    def __init__(self, collection, channel):

        self.cursor = None
        self.collection = collection
        self.channel = channel
        self.cursor = self.collection.find(
            {"channel": self.channel, "when": {"$gte": datetime.utcnow()}},
            cursor_type=CursorType.TAILABLE_AWAIT,
        )

    def __aiter__(self):
        return self

    async def __anext__(self):
        while self.cursor.alive:
            try:
                async for message in self.cursor:
                    if message["type"] == "message":
                        return message["message"]
            except StopAsyncIteration:
                await asyncio.sleep(0.5)


class MongoPubSub(PubSub):
    def __init__(
        self,
        host="localhost",
        port=27017,
        max_pool_size=100,
        database="anypubsub",
        collection_name="anyps_messages",
        collection_size=10 * 2 ** 20,
    ):
        if motor_installed is False:
            raise RuntimeError("Please install `motor`")  # pragma: no cover
        self.client = motor.motor_asyncio.AsyncIOMotorClient(
            host=host, port=port, maxPoolSize=max_pool_size
        )

        self.db = self.client[database]
        self.collection = None
        self.collection_name = collection_name
        self.collection_size = collection_size

    async def init(self):
        try:
            await self.db.create_collection(
                self.collection_name, size=self.collection_size, capped=True
            )
        except CollectionInvalid as e:
            logging.error("Error during create collection: {}".format(e))
        self.collection = self.db[self.collection_name]

    async def publish(self, channel, message):
        await self.collection.insert_one(
            {"type": "message", "channel": channel, "message": message, "when": datetime.utcnow()}
        )

    async def subscribe(self, channel):
        subscriber = MongoSubscriber(self.collection, channel)
        return subscriber



from setuptools import setup, find_packages
import os
from aio_pubsub._version import __version__

here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
except IOError:
    README = ''

setup(
    name='aio_pubsub',
    version=__version__,
    description="A generic interface wrapping multiple backends to provide a consistent pubsub API.",
    long_description=README,
    # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
    ],
    keywords='pubsub pub/sub mongodb',
    author='Bakhtiyor Ruziev',
    author_email='bakhtiyor.ruziev@yandex.ru',
    url='http://github.com/bruziev/async_pubsub',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    extras_require={
        'mongodb': [
            'motor >= 2',
        ],
        'aioredis': [
            'aioredis >= 1.2.0',
        ],

    },

)
